"""add last login

Revision ID: 4099fa344171
Revises: 139ec7ed17b7
Create Date: 2013-10-24 10:13:45.104902

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "4099fa344171"
down_revision = "139ec7ed17b7"


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column("persons", sa.Column("last_login", sa.DateTime(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("persons", "last_login")
    ### end Alembic commands ###



from tahrir_api.dbapi import TahrirDatabase

db = TahrirDatabase("backend://badges:badgesareawesome@localhost/badges")

badge_id = "fossbox"
person_email = "person@email.com"
issued_on = None

db.add_person(person_email)
db.add_assertion(badge_id, person_email, issued_on)



import importlib.resources
import os

from paste.deploy import appconfig
from sqlalchemy_helpers import DatabaseManager

from ..model import Assertion, Badge, DeclarativeBase, Issuer, Person  # noqa: F401


def _getpathsec(config_uri, name):
    if "#" in config_uri:
        path, section = config_uri.split("#", 1)
    else:
        path, section = config_uri, "main"
    if name:
        section = name
    return path, section


def get_db_manager_from_paste(config_uri):
    path, section = _getpathsec(config_uri, "pyramid")
    config_name = f"config:{path}"
    here_dir = os.getcwd()

    global_conf = None
    if "OPENSHIFT_APP_NAME" in os.environ:
        if "OPENSHIFT_MYSQL_DB_URL" in os.environ:
            template = "{OPENSHIFT_MYSQL_DB_URL}{OPENSHIFT_APP_NAME}"
        elif "OPENSHIFT_POSTGRESQL_DB_URL" in os.environ:
            template = "{OPENSHIFT_POSTGRESQL_DB_URL}{OPENSHIFT_APP_NAME}"

        global_conf = {"sqlalchemy.url": template.format(**os.environ)}

    settings = appconfig(config_name, name=section, relative_to=here_dir, global_conf=global_conf)

    prefix = "sqlalchemy."
    db_options = {key[len(prefix) :]: settings[key] for key in settings if key.startswith(prefix)}
    dburi = db_options.pop("url")
    with importlib.resources.as_file(
        importlib.resources.files("tahrir_api").joinpath("migrations")
    ) as alembic_path:
        return DatabaseManager(dburi, alembic_path.as_posix())



import pytest

from tahrir_api.dbapi import TahrirDatabase


@pytest.fixture
def callback_calls():
    return []


@pytest.fixture
def api(callback_calls, tmp_path):
    def callback(*args, **kwargs):
        callback_calls.append((args, kwargs))

    dbapi = TahrirDatabase(
        f"sqlite:///{tmp_path.as_posix()}/testdb.db", notification_callback=callback
    )
    dbapi.db_mgr.sync()
    return dbapi



import datetime
import hashlib
import time
import uuid

import arrow
import pygments
import simplejson
from sqlalchemy import Column, DateTime, ForeignKey, Unicode, UniqueConstraint
from sqlalchemy.orm import object_session, relationship, sessionmaker
from sqlalchemy.types import Boolean, Integer
from sqlalchemy_helpers import Base as DeclarativeBase


class Issuer(DeclarativeBase):
    __tablename__ = "issuers"
    id = Column(Integer, unique=True, primary_key=True)
    origin = Column(Unicode(128), nullable=False)
    name = Column(Unicode(128), nullable=False, unique=True)
    org = Column(Unicode(128), nullable=False)
    contact = Column(Unicode(128), nullable=False)
    badges = relationship("Badge", backref="issuer")
    created_on = Column(DateTime, nullable=False, default=datetime.datetime.now)

    def __str__(self):
        return str(self.name)

    def as_dict(self):
        return dict(
            origin=self.origin,
            name=self.name,
            org=self.org,
            contact=self.contact,
            created_on=time.mktime(self.created_on.timetuple()),
        )


def generate_default_id(context):
    return context.current_parameters["name"].lower().replace(" ", "-")


class Badge(DeclarativeBase):
    __tablename__ = "badges"
    id = Column(Unicode(128), primary_key=True, default=generate_default_id)
    name = Column(Unicode(128), nullable=False, unique=True)
    image = Column(Unicode(128), nullable=False)
    stl = Column(Unicode(128))
    description = Column(Unicode(128), nullable=False)
    criteria = Column(Unicode(128), nullable=False)
    issuer_id = Column(Integer, ForeignKey("issuers.id"), nullable=False)
    milestone = relationship("Milestone", backref="badge")
    authorizations = relationship("Authorization", backref="badge")
    assertions = relationship("Assertion", backref="badge")
    invitations = relationship("Invitation", backref="badge")
    created_on = Column(DateTime, nullable=False, default=datetime.datetime.now)
    tags = Column(Unicode(128))

    def __str__(self):
        return str(self.name)

    def as_dict(self):
        if self.image.startswith("http"):
            image = self.image
        else:
            image = "/pngs/" + self.image
        return dict(
            version="0.5.0",
            name=self.name,
            image=image,
            description=self.description,
            criteria=self.criteria,
            issuer=self.issuer.as_dict(),
            created_on=time.mktime(self.created_on.timetuple()),
            tags=self.tags,
        )

    def authorized(self, person):
        """Return true if a given person is authorized to admin a badge"""
        for authz in self.authorizations:
            if authz.person == person:
                return True

        return False


class Team(DeclarativeBase):
    __tablename__ = "team"
    id = Column(Unicode(128), primary_key=True, default=generate_default_id)
    name = Column(Unicode(128), nullable=False, unique=True)
    series = relationship("Series", backref="team")
    created_on = Column(DateTime, nullable=False, default=datetime.datetime.now)

    def as_dict(self):
        return dict(id=self.id, name=self.name, created_on=str(self.created_on))


class Series(DeclarativeBase):
    __tablename__ = "series"
    id = Column(Unicode(128), primary_key=True, default=generate_default_id)
    name = Column(Unicode(128), nullable=False, unique=True)
    description = Column(Unicode(128), nullable=False)
    created_on = Column(DateTime, nullable=False, default=datetime.datetime.now)
    last_updated = Column(
        DateTime,
        nullable=False,
        default=datetime.datetime.now,
        onupdate=datetime.datetime.now,
    )
    tags = Column(Unicode(128))
    milestone = relationship("Milestone", backref="series")
    team_id = Column(Unicode(128), ForeignKey("team.id"), nullable=False)

    def as_dict(self):
        return dict(
            id=self.id,
            name=self.name,
            created_on=str(self.created_on),
            last_updated=str(self.last_updated),
            team=self.team.as_dict(),
        )


class Milestone(DeclarativeBase):
    __tablename__ = "milestone"
    __table_args__ = (UniqueConstraint("position", "badge_id", "series_id"),)
    id = Column(Integer, unique=True, primary_key=True)
    position = Column(Integer, default=None)
    badge_id = Column(Unicode(128), ForeignKey("badges.id"), nullable=False)
    series_id = Column(Unicode(128), ForeignKey("series.id"), nullable=False)

    def as_dict(self):
        return dict(
            position=self.position,
            badge=self.badge.as_dict(),
            series_id=self.series.id,
        )


class Person(DeclarativeBase):
    __tablename__ = "persons"
    id = Column(Integer, unique=True, primary_key=True)
    email = Column(Unicode(128), nullable=False, unique=True)
    _avatar = Column(Unicode(128), nullable=True)
    authorizations = relationship("Authorization", backref="person")
    assertions = relationship("Assertion", backref="person")
    nickname = Column(Unicode(128), unique=True)
    website = Column(Unicode(128))
    bio = Column(Unicode(140))
    created_on = Column(DateTime, nullable=False, default=datetime.datetime.now)
    last_login = Column(DateTime, nullable=True, default=None)
    opt_out = Column(Boolean, nullable=False, default=False)
    # An integer that organizes the users by the number of
    # badges they have ever been awarded.  A value of None
    # indicates that they have not been ranked yet at all.
    rank = Column(Integer, default=None)

    def __repr__(self):
        return f"<Person: '{self.nickname} <{self.email}>'"

    @property
    def email_sha1(self):
        return hashlib.sha1(self.email.encode("utf-8")).hexdigest()

    @property
    def avatar(self):
        return self._avatar or self.email

    def __str__(self):
        return str(self.email)

    def as_dict(self):
        return dict(
            email=self.email,
            id=self.id,
            nickname=self.nickname,
            website=self.website,
            bio=self.bio,
            rank=self.rank,
        )


def invitation_id_default(context):
    return hashlib.md5(salt_default(context).encode("utf-8")).hexdigest()


class Invitation(DeclarativeBase):
    """This is a temporary invitation to receive a badge.

    The idea is that a user can create a "You made my day" badge, and
    then award it to another user.  However, instead of just directly
    associating the righthand user with the badge, we "invite" them
    to accept it.

    """

    __tablename__ = "invitations"
    id = Column(Unicode(32), primary_key=True, unique=True, default=invitation_id_default)
    created_on = Column(DateTime, nullable=False)
    expires_on = Column(DateTime, nullable=False)
    badge_id = Column(Unicode(128), ForeignKey("badges.id"), nullable=False)
    created_by = Column(Integer, ForeignKey("persons.id"), nullable=False)

    @property
    def expired(self):
        return datetime.datetime.now() > self.expires_on

    @property
    def expires_on_relative(self):
        return arrow.get(self.expires_on).humanize()


class Authorization(DeclarativeBase):
    __tablename__ = "authorizations"
    id = Column(Integer, primary_key=True)
    badge_id = Column(Unicode(128), ForeignKey("badges.id"), nullable=False)
    person_id = Column(Integer, ForeignKey("persons.id"), nullable=False)


def recipient_default(context):
    Session = sessionmaker(context.engine)()
    person_id = context.current_parameters["person_id"]
    person = Session.query(Person).filter_by(id=person_id).one()

    return hashlib.sha256(
        (person.email + context.current_parameters["salt"]).encode("utf-8")
    ).hexdigest()


def salt_default(context):
    return str(uuid.uuid4())


def assertion_id_default(context):
    person_id = context.current_parameters["person_id"]
    badge_id = context.current_parameters["badge_id"]
    return f"{badge_id} -> {person_id}"


class Assertion(DeclarativeBase):
    __tablename__ = "assertions"
    id = Column(Unicode(128), primary_key=True, unique=True, default=assertion_id_default)
    badge_id = Column(Unicode(128), ForeignKey("badges.id"), nullable=False)
    person_id = Column(Integer, ForeignKey("persons.id"), nullable=False)
    salt = Column(Unicode(128), nullable=False, default=salt_default)
    issued_on = Column(DateTime, nullable=False, default=datetime.datetime.now)

    # An optional link back to the event that warranted the award
    issued_for = Column(Unicode(256))

    recipient = Column(Unicode(256), nullable=False, default=recipient_default)

    def __str__(self):
        return str(self.badge) + "<->" + str(self.person)

    @property
    def _recipient(self):
        return f"sha256${self.recipient}"

    def as_dict(self):
        result = dict(recipient=self._recipient, salt=self.salt, badge=self.badge.as_dict())
        # Eliminate this check since I made issued_on not nullable?
        if self.issued_on:
            result["issued_on"] = self.issued_on.strftime("%Y-%m-%d")
        return result

    def __getitem__(self, key):
        if key not in ("pygments", "delete"):
            raise KeyError
        return getattr(self, f"__{key}__")()

    def __delete__(self):
        session = object_session(self)
        return lambda: session.delete(self)

    def __pygments__(self):
        html_args = {"full": False}
        pretty_encoder = simplejson.encoder.JSONEncoder(indent=2)
        html = pygments.highlight(
            pretty_encoder.encode(self.as_dict()),
            pygments.lexers.JavascriptLexer(),
            pygments.formatters.HtmlFormatter(**html_args),
        ).strip()
        return html



"""fix foreign key mismatch

Revision ID: 16943d9088cf
Revises: 24282792d72a
Create Date: 2013-06-23 22:55:47.775736

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "16943d9088cf"
down_revision = "24282792d72a"


def upgrade():
    op.alter_column("invitations", "created_by", type_=sa.Integer)


def downgrade():
    op.alter_column("invitations", "created_by", type_=sa.Unicode(128))



import os
import sys

from .utils import get_db_manager_from_paste


def usage(argv):
    cmd = os.path.basename(argv[0])
    print(f"usage: {cmd} <config_uri>\n '(example: \"{cmd} development.ini\"'")
    sys.exit(1)


def main(argv=sys.argv):
    if len(argv) != 2:
        usage(argv)

    config_uri = argv[1]
    db_mgr = get_db_manager_from_paste(config_uri)
    db_mgr.sync()



"""add authz table

Revision ID: ce541796a7
Revises: 4099fa344171
Create Date: 2013-12-13 15:30:16.576871

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "ce541796a7"
down_revision = "4099fa344171"


def upgrade():
    op.create_table(
        "authorizations",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("badge_id", sa.Unicode(length=128), nullable=False),
        sa.Column("person_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["badge_id"], ["badges.id"]),
        sa.ForeignKeyConstraint(["person_id"], ["persons.id"]),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade():
    op.drop_table("authorizations")



"""Add created_on field to issuers table.

Revision ID: fa1d309e8c3
Revises: 420c02357a1b
Create Date: 2013-06-10 12:30:31.850641

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "fa1d309e8c3"
down_revision = "420c02357a1b"


def upgrade():
    op.add_column("issuers", sa.Column("created_on", sa.DateTime, nullable=False))


def downgrade():
    # This will fail if using SQLite. The transaction will
    # be rolled back.
    op.drop_column("issuers", "created_on")



import click
from fasjson_client import Client as FasjsonClient
from fasjson_client.errors import APIError
from sqlalchemy import func, select

from ..model import Person
from .utils import get_db_manager_from_paste


@click.command()
@click.argument("paste-config", type=click.Path(exists=True))
@click.argument("fasjson-url", required=True)
def main(paste_config, fasjson_url):
    fasjson = FasjsonClient(url=fasjson_url)

    db_mgr = get_db_manager_from_paste(paste_config)
    with db_mgr.Session() as session:
        query = select(Person).where(
            Person.email.like("%@fedoraproject.org"),
            Person._avatar.is_(None),
            Person.opt_out.is_(False),
        )
        total = session.scalar(query.with_only_columns(func.count(Person.id)))
        click.echo(f"Found {total} users to update.")
        with click.progressbar(
            session.scalars(query),
            length=total,
            label="Setting avatar",
            item_show_func=lambda p: p.nickname if p else "",
        ) as persons:
            for person in persons:
                try:
                    fas_user = fasjson.get_user(username=person.nickname).result
                except APIError as e:
                    if e.code == 404:
                        click.echo(
                            f"User {person.nickname} ({person.email}) not found in FASJSON, "
                            "skipping"
                        )
                        continue
                    else:
                        raise
                person._avatar = fas_user["emails"][0]
                # print(f"Setting {person.nickname}'s avatar to {person._avatar}")
                session.commit()



""" Module to keep random utils. """


def autocommit(func):
    """A decorator that autocommits after API calls unless
    configured otherwise.
    """

    def _wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        if self.autocommit:
            self.session.commit()
        return result

    _wrapper.__name__ = func.__name__
    _wrapper.__doc__ = func.__doc__

    return _wrapper


def convert_name_to_id(name):
    """
    Convert a badge name into a valid badge ID.

    :type name: string
    :param name: The badge name to convert to an ID
    """

    badge_id = name.lower().replace(" ", "-")
    bad = ['"', "'", "(", ")", "*", "&", "?"]
    replacements = dict(zip(bad, [""] * len(bad)))
    for a, b in replacements.items():
        badge_id = badge_id.replace(a, b)

    return badge_id



"""Add two new columns for Person.

Revision ID: 3c7fd5b4e2c2
Revises: 24282792d72a
Create Date: 2013-06-26 14:46:28.361709

"""

import datetime

import sqlalchemy as sa
from alembic import op

import tahrir_api


# revision identifiers, used by Alembic.
revision = "3c7fd5b4e2c2"
down_revision = "16943d9088cf"


def upgrade():
    op.add_column("persons", sa.Column("created_on", sa.DateTime()))
    op.add_column("persons", sa.Column("opt_out", sa.Boolean()))

    # We have to do this manually because I can't figure out how to make
    # alembic apply defaults to sqlite.
    engine = op.get_bind().engine
    session_maker = sa.orm.sessionmaker(bind=engine)
    session = sa.orm.scoped_session(session_maker)
    persons = session.query(tahrir_api.model.Person).all()
    for person in persons:
        # Set our defaults
        person.created_on = datetime.datetime.now()
        person.opt_out = False
    session.commit()


def downgrade():
    op.drop_column("persons", "opt_out")
    op.drop_column("persons", "created_on")



"""issued_for

Revision ID: 2879ed5a6297
Revises: ce541796a7
Create Date: 2014-03-04 11:04:31.024949

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "2879ed5a6297"
down_revision = "ce541796a7"


def upgrade():
    op.add_column("assertions", sa.Column("issued_for", sa.Unicode(256)))


def downgrade():
    # This will fail if using SQLite. The transaction will
    # be rolled back.
    op.drop_column("assertions", "issued_for")



"""Add a rank column to the Person table.

Revision ID: 139ec7ed17b7
Revises: 3c7fd5b4e2c2
Create Date: 2013-08-16 12:06:00.092052

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "139ec7ed17b7"
down_revision = "3c7fd5b4e2c2"


def upgrade():
    op.add_column("persons", sa.Column("rank", sa.Integer, default=None))


def downgrade():
    op.drop_column("persons", "rank")
